# coding: utf-8

from django.db.backends.mysql import base, creation


class SphinxOperations(base.DatabaseOperations):
    compiler_module = "sphinxsearch.backend.sphinx.compiler"

    def fulltext_search_sql(self, field_name):
        """ Formats full-text search expression."""
        return 'MATCH (\'@%s "%%s"\')' % field_name

    def force_no_ordering(self):
        """ Fix unsupported syntax "ORDER BY NULL"."""
        return []


class SphinxCreation(creation.DatabaseCreation):

    def create_test_db(self, *args, **kwargs):
        # NOOP, test using regular sphinx database.
        if self.connection.settings_dict.get('TEST_NAME'):
            # initialize connection database name
            test_name = self.connection.settings_dict['TEST_NAME']
            self.connection.close()
            self.connection.settings_dict['NAME'] = test_name
            self.connection.cursor()
            return test_name
        return self.connection.settings_dict['NAME']

    def destroy_test_db(self, *args, **kwargs):
        # NOOP, we created nothing, nothing to destroy.
        return


class DatabaseWrapper(base.DatabaseWrapper):
    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)
        self.ops = SphinxOperations(self)
        self.creation = SphinxCreation(self)
        # The following can be useful for unit testing, with multiple databases
        # configured in Django, if one of them does not support transactions,
        # Django will fall back to using clear/create
        # (instead of begin...rollback) between each test. The method Django
        # uses to detect transactions uses CREATE TABLE and DROP TABLE,
        # which ARE NOT supported by Sphinx, even though transactions ARE.
        # Therefore, we can just set this to True, and Django will use
        # transactions for clearing data between tests when all OTHER backends
        # support it.
        self.features.supports_transactions = True
        self.features.allows_group_by_pk = False
        self.features.uses_savepoints = False
        self.features.supports_column_check_constraints = False



# coding: utf-8

# $Id: $
from collections import OrderedDict
import functools
from django.db import models
from django.db.models import Count, BooleanField
from django.db.models.base import ModelBase
from django.db.models.expressions import Col, Func, BaseExpression
from django.db.models.sql import Query
from django.db.models.sql.where import WhereNode, ExtraWhere
from django.utils.datastructures import OrderedSet


class SphinxCount(Count):
    """ Replaces Mysql-like COUNT('*') with COUNT(*) token."""
    template = '%(function)s(*)'

    def as_sql(self, compiler, connection, function=None, template=None):
        sql, params = super(SphinxCount, self).as_sql(
            compiler, connection, function=function, template=template)
        try:
            params.remove('*')
        except ValueError:
            pass
        return sql, params


class SphinxWhereExpression(BaseExpression):
    def __init__(self, where, where_params):
        self.where = where
        self.where_params = where_params
        super(SphinxWhereExpression, self).__init__(output_field=BooleanField())

    def as_sql(self, compiler, connection):
        return self.where, self.where_params


class SphinxExtraWhere(ExtraWhere):

    def as_sql(self, qn=None, connection=None):
        sqls = ["%s" % sql for sql in self.sqls]
        return " AND ".join(sqls), tuple(self.params or ())


class SphinxWhereNode(WhereNode):

    def make_atom(self, child, qn, connection):
        """
        Transform search, the keyword should not be quoted.
        """
        return super(WhereNode, self).make_atom(child, qn, connection)
        lvalue, lookup_type, value_annot, params_or_value = child
        sql, params = super(SphinxWhereNode, self).make_atom(child, qn, connection)
        if lookup_type == 'search':
            if hasattr(lvalue, 'process'):
                try:
                    lvalue, params = lvalue.process(lookup_type, params_or_value, connection)
                except EmptyShortCircuit:
                    raise EmptyResultSet
            if isinstance(lvalue, tuple):
                # A direct database column lookup.
                field_sql = self.sql_for_columns(lvalue, qn, connection)
            else:
                # A smart object with an as_sql() method.
                field_sql = lvalue.as_sql(qn, connection)
            # TODO: There are a couple problems here.
            # 1. The user _might_ want to search only a specific field.
            # 2. However, since Django requires a field name to use the __search operator
            #    There is no way to do a search in _all_ fields.
            # 3. Because, using multiple __search operators is not supported.
            # So, we need to merge multiped __search operators into a single MATCH(), we
            # can't do that here, we have to do that one level up...
            # Ignore the field name, search all fields:
            params = ('@* %s' % params[0], )
            # _OR_ respect the field name, and search on it:
            #params = ('@%s %s' % (field_sql, params[0]), )
        if self._real_negated:
            col = lvalue.col
            if lookup_type == 'exact':
                sql = '%s <> %%s' % col
            elif lookup_type == 'in':
                params_placeholder = '(%s)' % (', '.join(['%s'] * len(params)))
                sql = '%s NOT IN %s' % (col, params_placeholder)
            else:
                raise ValueError("Negative '%s' lookup not supported" % lookup_type)
        return sql, params


class SphinxQuery(Query):
    _clonable = ('options', 'match', 'group_limit', 'group_order_by',
                 'with_meta')

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('where', SphinxWhereNode)
        super(SphinxQuery, self).__init__(*args, **kwargs)

    def clone(self, klass=None, memo=None, **kwargs):
        query = super(SphinxQuery, self).clone(klass=klass, memo=memo, **kwargs)
        for attr_name in self._clonable:
            value = getattr(self, attr_name, None)
            if value:
                setattr(query, attr_name, value)
        return query

    def add_match(self, *args, **kwargs):
        if not hasattr(self, 'match'):
            self.match = OrderedDict()
        for expression in args:
            self.match.setdefault('*', OrderedSet())
            if isinstance(expression, (list, tuple)):
                self.match['*'].update(expression)
            else:
                self.match['*'].add(expression)
        for field, expression in kwargs.items():
            self.match.setdefault(field, OrderedSet())
            if isinstance(expression, (list, tuple, set)):
                self.match[field].update(expression)
            else:
                self.match[field].add(expression)

    def get_count(self, using):
        """
        Performs a COUNT() query using the current filter constraints.
        """
        obj = self.clone()
        obj.add_annotation(SphinxCount('*'), alias='__count', is_summary=True)
        number = obj.get_aggregation(using, ['__count'])['__count']
        if number is None:
            number = 0
        return number


class SphinxCol(Col):
    def as_sql(self, compiler, connection):
        # As column names in SphinxQL couldn't be escaped with `backticks`,
        # simply return column name
        return self.target.column, []


class SphinxModelBase(ModelBase):

    def __new__(cls, name, bases, attrs):
        # Each field must be monkey-patched with SphinxCol class to prevent
        # `tablename`.`attr` appearing in SQL
        for attr in attrs.values():
            if isinstance(attr, models.Field):
                col_patched = getattr(attr, '_col_patched', False)
                if not col_patched:
                    cls.patch_col_class(attr)

        new_class = super(SphinxModelBase, cls).__new__(cls, name, bases, attrs)

        # if have overriden primary key, it should be the first local field,
        # because of JSONField feature at jsonfield.subclassing.Creator.__set__
        local_fields = new_class._meta.local_fields
        try:
            pk_idx = local_fields.index(new_class._meta.pk)
            if pk_idx > 0:
                local_fields.insert(0, local_fields.pop(pk_idx))
        except ValueError:
            pass

        return new_class

    def add_to_class(cls, name, value):
        col_patched = getattr(value, '_col_patched', False)
        if not col_patched and isinstance(value, models.Field):
            cls.patch_col_class(value)
        super(SphinxModelBase, cls).add_to_class(name, value)

    @classmethod
    def patch_col_class(cls, field):
        @functools.wraps(field.get_col)
        def wrapper(alias, output_field=None):
            col = models.Field.get_col(field, alias, output_field=output_field)
            col.__class__ = SphinxCol
            return col
        field.get_col = wrapper



# coding: utf-8
from copy import copy

from django.conf import settings
from django.db import connections
from django.db.models import QuerySet
from django.db.models.expressions import RawSQL

from sphinxsearch import sql
from sphinxsearch.fields import *
from sphinxsearch.utils import sphinx_escape


class SphinxQuerySet(QuerySet):
    def __init__(self, model, **kwargs):
        kwargs.setdefault('query', sql.SphinxQuery(model))
        super(SphinxQuerySet, self).__init__(model, **kwargs)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        args = list(args)
        kwargs = copy(kwargs)
        for key, value in list(kwargs.items()):
            field, lookup = self.__get_field_lookup(key)
            if self.__check_search_lookup(field, lookup, value):
                kwargs.pop(key, None)
            elif self.__check_in_lookup(field, lookup, value, negate):
                kwargs.pop(key, None)
            elif self.__check_sphinx_field_exact(field, lookup, value, negate):
                kwargs.pop(key, None)
            elif self.__check_mva_field_lookup(field, lookup, value, negate):
                kwargs.pop(key, None)
            pass

        return super(SphinxQuerySet, self)._filter_or_exclude(negate, *args,
                                                              **kwargs)

    def __get_field_lookup(self, key):
        tokens = key.split('__')
        if len(tokens) == 1:
            field_name, lookup = tokens[0], 'exact'
        elif len(tokens) == 2:
            field_name, lookup = tokens
        else:
            raise ValueError("Nested field lookup found")
        if field_name == 'pk':
            field = self.model._meta.pk
        else:
            field = self.model._meta.get_field(field_name)
        return field, lookup

    def _negate_expression(self, negate, lookup):
        if isinstance(lookup, (tuple, list)):
            result = []
            for v in lookup:
                result.append(self._negate_expression(negate, v))
            return result
        else:
            if not isinstance(lookup, six.string_types):
                lookup = six.text_type(lookup)

            if not lookup.startswith('"'):
                lookup = '"%s"' % lookup
            if negate:
                lookup = '-%s' % lookup
            return lookup

    def match(self, *args, **kwargs):
        """ Enables full-text searching in sphinx (MATCH expression).

        qs.match('sphinx_expression_1', 'sphinx_expression_2')
            compiles to
        MATCH('sphinx_expression_1 sphinx_expression_2)

        qs.match(field1='sphinx_loopup1',field2='sphinx_loopup2')
            compiles to
        MATCH('@field1 sphinx_lookup1 @field2 sphinx_lookup2')
        """
        qs = self._clone()
        qs.query.add_match(*args, **kwargs)
        return qs

    def options(self, **kw):
        """ Setup OPTION clause for query."""
        qs = self._clone()
        try:
            qs.query.options.update(kw)
        except AttributeError:
            qs.query.options = kw
        return qs

    def with_meta(self):
        """ Force call SHOW META after fetching queryset data from searchd."""
        qs = self._clone()
        qs.query.with_meta = True
        return qs

    def group_by(self, *args, **kwargs):
        """
        :param args: list of fields to group by
        :type args: list-like

        Keyword params:
        :param group_limit: (GROUP <N> BY), limits number of group members to N
        :type group_limit: int
        :param group_order_by: (WITHIN GROUP ORDER BY), sort order within group
        :type group_order_by: list-like
        :return: new queryset with grouping
        :rtype: SphinxQuerySet
        """
        group_limit = kwargs.get('group_limit', 0)
        group_order_by = kwargs.get('group_order_by', ())
        if not isinstance(group_order_by, (list, tuple)):
            group_order_by = [group_order_by]
        qs = self._clone()
        qs.query.group_by = qs.query.group_by or []
        for field_name in args:
            if field_name not in qs.query.extra_select:
                field = self.model._meta.get_field_by_name(field_name)[0]
                qs.query.group_by.append(field.attname)
            else:
                qs.query.group_by.append(RawSQL(field_name, []))
        qs.query.group_limit = group_limit
        qs.query.group_order_by = group_order_by
        return qs

    def __check_mva_field_lookup(self, field, lookup, value, negate):
        """ Replaces some MVA field lookups with valid sphinx expressions."""

        if not isinstance(field, (SphinxMultiField, SphinxMulti64Field)):
            return False

        transforms = {
            'exact': "IN(%s, %%s)",
            'gte': "LEAST(%s) >= %%s",
            'ge': "LEAST(%s) > %%s",
            'lt': "GREATEST(%s) < %%s",
            'lte': "GREATEST(%s) <= %%s"
        }

        if lookup in transforms.keys():
            tpl = transforms[lookup]
            condition = tpl % field.column
            if negate:
                condition = "NOT (%s)" % condition
            self.query.add_extra(None, None, [condition], [value], None, None)
            return True
        else:
            raise ValueError("Invalid lookup for MVA: %s" % lookup)

    def __check_search_lookup(self, field, lookup, value):
        """ Replaces field__search lookup with MATCH() call."""
        if lookup != 'search':
            return False
        self.query.add_match(**{field.name: sphinx_escape(value)})
        return True

    def __check_in_lookup(self, field, lookup, value, negate):
        if lookup != 'in':
            return False
        if not isinstance(value, (tuple, list)):
            value = [value]
        placeholders = ', '.join(['%s'] * len(value))
        condition = "IN(%s, %s)" % (field.column, placeholders)
        value = [field.get_prep_value(v) for v in value]
        if negate:
            condition = "NOT (%s)" % condition
        self.query.add_extra(None, None, [condition], value, None, None)
        return True

    def __check_sphinx_field_exact(self, field, lookup, value, negate):
        if not isinstance(field, SphinxField):
            return False
        if lookup != 'exact':
            raise ValueError("Unsupported lookup for SphinxField")
        if negate:
            value = '-%s' % value
        self.query.add_match(**{field.column: value})
        return True

    def _fetch_meta(self):
        c = connections[settings.SPHINX_DATABASE_NAME].cursor()
        try:
            c.execute("SHOW META")
            self.meta = dict([c.fetchone()])
        except UnicodeDecodeError:
            self.meta = {}
        finally:
            c.close()

    def iterator(self):
        for row in super(SphinxQuerySet, self).iterator():
            yield row
        if getattr(self.query, 'with_meta', False):
            self._fetch_meta()

class SphinxManager(models.Manager):
    use_for_related_fields = True

    def get_queryset(self):
        """ Creates new queryset for model.

        :return: model queryset
        :rtype: SphinxQuerySet
        """

        # Determine which fields are sphinx fields (full-text data) and
        # defer loading them. Sphinx won't return them.
        # TODO: we probably need a way to keep these from being loaded
        # later if the attr is accessed.
        sphinx_fields = [field.name for field in self.model._meta.fields
                         if isinstance(field, SphinxField)]

        return SphinxQuerySet(self.model).defer(*sphinx_fields)

    def options(self, **kw):
        return self.get_queryset().options(**kw)

    def match(self, expression):
        return self.get_queryset().match(expression)

    def group_by(self, *args, **kw):
        return self.get_queryset().group_by(*args, **kw)

    def get(self, *args, **kw):
        return self.get_queryset().get(*args, **kw)


class SphinxModel(six.with_metaclass(sql.SphinxModelBase, models.Model)):
    class Meta:
        abstract = True

    objects = SphinxManager()

    _excluded_update_fields = (
        models.CharField,
        models.TextField
    )
